--[[
Patch by sarf:

Added Get_Debuffs function (retrieves all debuffs in an array, in either ['Poison'], ['Disease'], ['Magic'], ['Curse'])
 this is used in CureUnit (since it now checks for magic/disease/poison)

Added disease/magic handling (CureUnit)
Added Remove Lesser Curse to Mage and handling in the uncurse function
Added Paladin spells
Added some (all?) Priest racial buff spells (Dwarf, Night Elf, Troll, I think)

Fixed taking [class]['Buffs'][<buff>] == 1 into consideration.

Modified Thorns Rank 2 to use 6 as minimum level.


Suggestion:
Move all string constants to a seperate file (localization.lua) to make it possible to localize the thingy.
 This would allow translation to German/French/Swahili if necessary.

Add checking for number of negative effects when using Dispel Magic (rank 2 removes 2 effects).

Add checking for SpellBomb and DO NOT CAST ANY SPELLS WHILE IT IS AFFECTING THE USER.


]]--

--[[---------------------------------------------------------------------------
BuffBot 0.16
By Trimble Epic

This is my BuffBot mod.


todo:
Add 'sensor' to let the player know that some buffs can be recast.  we don't want to have to constantly
	tap the buffbot key every minute just to see if something needs to be done.
	this will require an OnUpdate event.
Add an override system - if a player specifically requests a prefered buff to be cast or not cast on
	them, I want to be able to use a slash command (and UI later) so that the buffer can specify
	that the target prefers something.  I.E. - warlock requests to get thorns, caster can target the 
	warlock and then do /buffbot prefers thorns, or target a druid healer and do
	/buffbot prefers no Thorns
	This information needs to be saved acorss sessions (## SavedVariable)
Find a way to buff party pets - perhaps monitor chat for pet buffs/debuffs?
add blacklist system for when you encounter a jerk and want to remember to 'never buff this asshole
	again.  include switch for only if not in party, and always.  should be command line driven to
	keep it hidden.
	This information needs to be saved acorss sessions (## SavedVariable)

Version History:
0.16	Mar 24, 2005
		Testing to work with WoW client version 1300
		Updated TOC to version 1300
0.15	Feb xx, 2005
		Added PRIEST support!
		Total rewrite of RangeToUnit function
			Resolved the problem of freezes/lag in instances
			Newly recalibrated
		Combined: JustCure function now handles cure poison and cure curse.
			removed function BuffBot_JustUncurse()
			removed key binding for uncurse
0.14	Feb 22, 2005
		Testing to work with WoW client version 4216
		Updated TOC to version 4216
		Omen of clarity druid self buff disabled, pending testing
0.13	Feb 19, 2005
		Bug fixes
			fixed bug when using classes other than druids.
		Added systems:
			Self Buffs
				Buffbot now correctly handles self buffs
			Redo - forces a full rebuff of the party by resetting the timers to make
				buffbot think all buffs are about to expire.
			Greater/Lesser buffs
				some buffs are lesser or greater versions of other buffs with different
				names.  Needed to parse all version to select the most powerful buff.
		Added new class buff - Mage


-------------------------"RlrlrlRLRLLRLGRLGRLGlglglglgl!"--------------------]]



--[[---------------------------------------------------------------------------
USER CONFIGURABLE SETTINGS

these are things you can freely adjust.
-----------------------------------------------------------------------------]]

--Note that these are PERCENT values, so remember the decimal point.
local min_health_to_buff    = .75
local min_health_to_cure    = .35
local min_health_to_uncurse = .50

--[[---------------------------------------------------------------------------
Local declarations  - LEAVE THESE ALONE :)
-----------------------------------------------------------------------------]]
local _version = '0.16'
local _testing = false
local _class
local _OnEvent = {}
local party_buffs = {}
local buffs_cast = {}
local spell_ids = {}
local party_ids = {player = 0, party1 = 1, party2 = 2, party3 = 3, party4 = 4, pet = 5}
local In_Instance

local spell_predata = {
	['Druid'] = {
		['Buffs'] = {
			['Thorns'] = 1,
			['Mark of the Wild'] = 1,
		},
		['Self Buffs'] = {
		},
		['Mark of the Wild'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 10,
			['Rank 4'] = 20,
			['Rank 5'] = 30,
			['Rank 6'] = 40,
			['Rank 7'] = 50,
			['Classes to Avoid'] = {},
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
			['Lesser of'] = 'Gift of the Wild',
		},
		['Thorns'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 6,
			['Rank 3'] = 14,
			['Rank 4'] = 24,
			['Rank 5'] = 34,
			['Rank 6'] = 44,
			['Classes to Avoid'] = {
				['Mage'] = 1,
				['Warlock'] = 1,
				['Priest'] = 1,
			},
			['Duration'] = 10 * 60,
			['Refresh'] = 2 * 60,
		},
		['Cure Poison'] = {
			['Lesser of'] = 'Abolish Poison',
			['x'] = 1,
		},
		['Abolish Poison'] = {
			['Greater of'] = 'Cure Poison',
			['x'] = 16,
		},
		['Remove Curse'] = {
			['x'] = 1,
		},
	},
	
	['Mage'] = {
		['Buffs'] = {
			['Arcane Intellect'] = 1,
		},
		['Self Buffs'] = {
			--['Mana Shield'] = 1,
			['Dampen Magic'] = 1,
			['Ice Armor'] = 1,
			['Frost Armor'] = 1,
		},
		['Arcane Intellect'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 4,
			['Rank 3'] = 18,
			['Rank 4'] = 32,
			['Rank 5'] = 46,
			['Classes to Avoid'] = {
				['Rogue'] = 1,
				['Warrior'] = 1,
			},
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
		},
		['Frost Armor'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
			['Lesser of'] = 'Ice Armor',
		},
		['Ice Armor'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Rank 4'] = 1,
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
			['Greater of'] = 'Frost Armor',
		},
		['Mana Shield'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Rank 4'] = 1,
			['Rank 5'] = 1,
			['Rank 6'] = 1,
			['Duration'] = 60,
			['Refresh'] = 7,
		},
		['Dampen Magic'] = {
			['Rank 1'] = 2,
			['Rank 2'] = 14,
			['Rank 3'] = 26,
			['Rank 4'] = 38,
			['Rank 5'] = 50,
			['Classes to Avoid'] = {},
			['Duration'] = 3 * 60,
			['Refresh'] = 30,
			['Lesser of'] = 'Amplify Magic',
		},
		['Amplify Magic'] = {  -- not used by buffbot at this time
			['Rank 1'] = 8,
			['Rank 2'] = 20,
			['Rank 3'] = 32,
			['Rank 4'] = 44,
			['Classes to Avoid'] = {},
			['Duration'] = 3 * 60,
			['Refresh'] = 30,
			['Lesser of'] = 'Dampen Magic',
		},
		['Remove Lesser Curse'] = {
			['x'] = 1,
		},
	},
	
	['Warlock'] = {
		['Buffs'] = {},
		['Self Buffs'] = {
			['Demon Armor'] = 1,
			['Demon Skin'] = 1,
		},
		['Demon Skin'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
			['Lesser Of'] = 'Demon Armor',
		},
		['Demon Armor'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Rank 4'] = 1,
			['Rank 5'] = 1,
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
			['Greater Of'] = 'Demon Skin',
		},
	},
	
	['Priest'] = {
		['Buffs'] = {
			['Power Word: Fortitude'] = 1,
			['Divine Spirit'] = 1,
			['Shadow Protection'] = 1,
			['Fear Ward'] = 1,
		},
		['Self Buffs'] = {
			['Inner Fire'] = 1,
			['Shadowguard'] = 1,
			["Elune's Grace"] = 1,
		},
		['Inner Fire'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Rank 4'] = 1,
			['Rank 5'] = 1,
			['Rank 6'] = 1,
			['Duration'] = 3 * 60,
			['Refresh'] = 30,
		},
		['Shadowguard'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Rank 4'] = 1,
			['Rank 5'] = 1,
			['Rank 6'] = 1,
			['Classes to Avoid'] = {},
			['Duration'] = 10 * 60,
			['Refresh'] = 30,
		},
		["Elune's Grace"] = {
			['Rank 1'] = 1,
			['Rank 2'] = 1,
			['Rank 3'] = 1,
			['Rank 4'] = 1,
			['Rank 5'] = 1,
			['Classes to Avoid'] = {},
			['Duration'] = 3 * 60,
			['Refresh'] = 30,
		},
		["Fear Ward"] = {
			['Rank 1'] = 1,
			['Classes to Avoid'] = {},
			['Duration'] = 10 * 60,
			['Refresh'] = 30,
		},
		['Power Word: Fortitude'] = {
			['Rank 1'] = 1,
			['Rank 2'] = 10,
			['Rank 3'] = 20,
			['Rank 4'] = 30,
			['Rank 5'] = 40,
			['Rank 6'] = 50,
			['Classes to Avoid'] = {},
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
			['Lesser of'] = 'Prayer of Fortitude',
		},
		['Divine Spirit'] = {
			['Rank 1'] = 30,
			['Rank 2'] = 40,
			['Rank 3'] = 50,
			['Classes to Avoid'] = {},
			['Duration'] = 30 * 60,
			['Refresh'] = 4 * 60,
		},
		['Shadow Protection'] = {
			['Rank 1'] = 20,
			['Rank 2'] = 30,
			['Rank 3'] = 40,
			['Classes to Avoid'] = {},
			['Duration'] = 10 * 60,
			['Refresh'] = 60,
		},
		['Cure Disease'] = {
			['Lesser of'] = 'Abolish Disease',
			['x'] = 1,
		},
		['Abolish Disease'] = {
			['Greater of'] = 'Cure Disease',
			['x'] = 16,
		},
		['Dispel Magic'] = {
			['x'] = 1,
		},
	},
	['Paladin'] = {
		['Purify'] = {
			['Lesser of'] = 'Cleanse',
			['x'] = 1,
		},
		['Cleanse'] = {
			['Greater of'] = 'Purify',
			['x'] = 16,
		},
	},
}

-- change this to 841 after changing RangeToUnit to not use sqrt... 
-- not that I think there's much of a performance boost, but hey, every bit helps, right? :)
BUFFBOT_EXTREME_RANGE = 29;

--[[---------------------------------------------------------------------------
Util functions
-----------------------------------------------------------------------------]]

local function Print(msg) 
	
	if not DEFAULT_CHAT_FRAME then return end
	
	DEFAULT_CHAT_FRAME:AddMessage(msg)
	
end

local function BigPrint(msg, color)
	
	local r = 1
	local g = 1
	local b = 1
	
	if color == 'green' then r = 0; b = 0 end
	if color == 'blue' then r = 0; g = 0 end
	if color == 'red' then g = 0; b = 0;r = 0.8 end
	if color == 'purple' then g = 0 end
	
	if not UIErrorsFrame then return end
	
	UIErrorsFrame:AddMessage(msg, r, g, b, 1.0, UIERRORS_HOLD_TIME)
	
end

local function Debug(...)
	
	if not _testing then return end
	
	local msg = 'BuffBot: '
	
	for k,v in ipairs(arg) do
		
		msg = msg .. tostring(v)
		
	end
	
	Print(msg)
	
end


--[[---------------------------------------------------------------------------
Helper Functions
-----------------------------------------------------------------------------]]

function RangeToUnit(target_unit)
	
	local px, py = GetPlayerMapPosition("player")
	
	if (px == 0 and py == 0) then
		
		if In_Instance then return -1 end
		
		SetMapToCurrentZone()
		
		if GetCurrentMapContinent() == 0 then
			
			In_Instance = true
			return -1
			
		else
			
			px, py = GetPlayerMapPosition("player")
			
		end
		
	else
		
		if In_Instance then 
			
			In_Instance = not In_Instance;
			SetMapToCurrentZone()
			
		end
		
	end
	
	local continent = GetCurrentMapContinent()
	local tx, ty = GetPlayerMapPosition(target_unit)
	
	if (tx == 0 and ty == 0) or (px == 0 and py == 0) then return -1 end
	
	if GetCurrentMapZone() > 0 then SetMapZoom(continent) end
	
	local mapFileName, textureHeight, textureWidth = GetMapInfo()
	local xdelta = ((tx-px) * textureWidth)   * 1.47167
	local ydelta = ((ty-py) * textureHeight)
	local distance
	
	if  continent == 2 then
		
		distance = sqrt(xdelta*xdelta + ydelta*ydelta) * 11.65
		
	elseif continent == 1 then
		
		distance = sqrt(xdelta*xdelta + ydelta*ydelta) * 12.13
		
	else
		
		return -1
		
	end
	
	return (math.floor(distance*100)/100)
	
end

local function UnitCanBeBuffed(unit)
	
	if not UnitExists(unit) then
		
		Debug(unit, ' does not exist')
		return false
		
	end
	
	if not UnitIsFriend('player', unit) then
		Debug(unit, ' is not a friend')
		return false
	end
	
	if UnitIsEnemy(unit, 'player') then
		Debug(unit, ' is an enemy')
		return false
	end
	
	if UnitCanAttack(unit, 'player') then
		Debug(unit, ' can attack player')
		return false
	end
	
	if UnitCanAttack('player', unit) then
		Debug(unit, ' can be attacked by player')
		return false
	end
	
	if not unit == 'player' and not UnitCanCooperate('player', unit) then
		Debug(unit, ' cannot cooperate with player')
		return false
	end
	
	if not UnitIsConnected(unit) then
		Debug(unit, ' is not connected')
		return false
	end
	
	if UnitIsDead(unit) then
		Debug(unit, ' is dead')
		return false
	end
	
	if UnitIsGhost(unit) then
		Debug(unit, ' is ghost')
		return false
	end
	
	if RangeToUnit(unit) > BUFFBOT_EXTREME_RANGE then
		Debug(unit, ' is out of range')
		return false
	end
	
	Debug(unit, ' is buffable')
	return true
	
end

local function LowestUnitHealth()
	
	local lowest_health = 100
	local unit_health = 100
	
	if UnitCanBeBuffed('target') then
		
		if UnitInParty('target') then
			
			return UnitHealth('target') / UnitHealthMax('target')
			
		else
			
			return UnitHealth('target')
			
		end
		
	else
		
		for unit in party_ids do
			
			unit_health = UnitHealth(unit) / UnitHealthMax(unit)
			
			if unit_health < lowest_health then lowest_health = unit_health end
			
		end
		
		return lowest_health
		
	end
	
end

local function Load_Spell_Ids()
	
	local i = 1
	
	spell_ids = {}
	
	while true do
		
		local spell_name, spell_rank = GetSpellName(i, SpellBookFrame.bookType)
		
		if not spell_name then break end
		if not spell_ids[spell_name] then spell_ids[spell_name] = {} end
		if spell_rank == "" then spell_rank = 'x' end
		spell_ids[spell_name][spell_rank] = i
		i = i + 1
		
		if i > 400 then break end -- safety
		
	end
	
end

local Spell_ID_By_Level
function Spell_ID_By_Level(spell_name, unit_level)
	
	if (not spell_ids[spell_name])
	or (not spell_predata[_class]) then return end
	
	local ranks = {'Rank 7','Rank 6','Rank 5','Rank 4','Rank 3','Rank 2','Rank 1','x'}
	
	for i,rank in ipairs(ranks) do
		
		if spell_ids[spell_name][rank] then
			
			if unit_level >= spell_predata[_class][spell_name][rank] then
				
				return spell_ids[spell_name][rank]
				
			end
			
		end
		
	end
	
	if  spell_predata[_class][spell_name]
	and spell_predata[_class][spell_name]['Greater of'] then
		
		
		lesser_spell = spell_predata[_class][spell_name]['Greater of']
		
		local lesser_spell_id = Spell_ID_By_Level(lesser_spell, unit_level)
		
		if lesser_spell_id then
			
			return lesser_spell_id
			
		end
		
	end
	
end

--[[---------------------------------------------------------------------------
Sense Debuffs
-----------------------------------------------------------------------------]]
local Find_Buff
function Find_Buff(_buff_to_find, unit)
	
	if spell_predata[_class][_buff_to_find]
	and spell_predata[_class][_buff_to_find]['Lesser Of'] then
		
		if Find_Buff(spell_predata[_class][_buff_to_find]['Lesser Of'],unit) then
			
			return true
			
		end
		
	end
	
	local buff_index = 1
	
	while UnitBuff(unit, buff_index) do
		
		BuffBotTooltip:SetUnitBuff(unit,buff_index)
		local buff_name = BuffBotTooltipTextLeft1:GetText()
		
		if buff_name == _buff_to_find then 
			
			return true 
			
		end
		
		buff_index = buff_index + 1
		
		if buff_index > 50 then break end -- saftey
		
	end
	
end

-- added by sarf
local function Get_Debuffs(unit)
	
	local debuffs = {};
	
	local debuff_index = 1
	
	while UnitDebuff(unit, debuff_index) do
		
		BuffBotTooltipTextLeft1:SetText(nil)
		BuffBotTooltipTextRight1:SetText(nil)
		BuffBotTooltip:SetUnitDebuff(unit,debuff_index)
		
		local debuff_name = BuffBotTooltipTextLeft1:GetText();
		local debuff_type = BuffBotTooltipTextRight1:GetText();
		
		if ( debuff_type ) then
			if ( not debuffs[debuff_type] ) then
				debuffs[debuff_type] = {};
			end
			if ( not debuff_name ) then
				debuff_name = "<NIL>";
			end
			table.insert(debuffs[debuff_type], debuff_name);
		end
		
		debuff_index = debuff_index + 1
		
		if debuff_index > 50 then break end -- saftey
		
	end
	
	return debuffs;
	
end

local function Find_Debuff(_debuff_to_find, unit)
	
	local debuff_index = 1
	
	while UnitDebuff(unit, debuff_index) do
		
		BuffBotTooltipTextRight1:SetText(nil)
		BuffBotTooltip:SetUnitDebuff(unit,debuff_index)
		
		local debuff_type = BuffBotTooltipTextRight1:GetText()
		
		if debuff_type == _debuff_to_find then return true end
		
		debuff_index = debuff_index + 1
		
		if debuff_index > 50 then break end -- saftey
		
	end
	
end

local function Load_Unit_Buffs(unit)
	
	if not UnitExists(unit) then return end
	
	if party_ids[unit] then
		
		party_buffs[unit] = {
			buffs   = {},
			debuffs = {},
			}
		
		local buff_index = 1
		
		while UnitBuff(unit, buff_index) do
			
			BuffBotTooltip:SetUnitBuff(unit,buff_index)
			local buff_name = BuffBotTooltipTextLeft1:GetText()
			
			party_buffs[unit]['buffs'][buff_name] = true
			buff_index = buff_index + 1
			
			if buff_index > 50 then break end -- saftey
			
		end
		
		local debuff_index = 1
		local debuff_type = ''
		
		while UnitDebuff(unit, debuff_index) do
			
			BuffBotTooltipTextRight1:SetText(nil)
			BuffBotTooltip:SetUnitDebuff(unit,debuff_index)
			local debuff_name = BuffBotTooltipTextLeft1:GetText()
			local debuff_type = BuffBotTooltipTextRight1:GetText()
			if debuff_type == nil then debuff_type = 'generic' end
			party_buffs[unit]['debuffs'][debuff_type] = true
			debuff_index = debuff_index + 1
			
			if debuff_index > 50 then break end -- saftey
			
		end
		
	end
	
end

local function Load_Party_Buffs()
	
	for unit in party_ids do
		
		if UnitExists(unit) then
			
			Load_Unit_Buffs(unit)
			
		else
			
			party_buffs[unit] = nil
			
		end
		
	end
	
end

local function Detect_Debuffs()
	
	for unit in party_ids do
		
		if party_buffs[unit] 
		and UnitExists(unit)then
			
			local debuffs = Get_Debuffs(unit);
			if unit == 'player' then
				if debuffs['Poison'] then BigPrint('You have been poisoned!','green') end
				if debuffs['Curse'] then BigPrint('You have been cursed!','red') end
				if debuffs['Magic'] then BigPrint('Harmful magic surrounds you!','blue') end
				if debuffs['Disease'] then BigPrint('You have been diseased!','purple') end
				
			else
				
				if debuffs['Poison'] then  BigPrint('You sense that '..UnitName(unit)..' has been poisoned!','green') end
				if debuffs['Curse'] then BigPrint('You sense a curse on '..UnitName(unit)..'!','red') end
				if debuffs['Magic'] then BigPrint('You sense harmful magic around '..UnitName(unit)..'!','blue') end
				if debuffs['Disease'] then BigPrint('You sense that '..UnitName(unit)..' has contracted a disease.','purple') end
				
			end
			
		end
		
	end
	
end

--[[---------------------------------------------------------------------------
BuffBot Internals
-----------------------------------------------------------------------------]]

local function CureUnit(unit)
	
	if LowestUnitHealth() < min_health_to_cure then return end
	
	if UnitCanBeBuffed(unit) 
	and RangeToUnit(unit) < BUFFBOT_EXTREME_RANGE then
	
		local debuffs = Get_Debuffs(unit);
		if ( debuffs['Poison'] ) then
			Debug('CureUnit: Poison Found! Attempting to cure')
			
			CurePoison_id = Spell_ID_By_Level('Abolish Poison',UnitLevel(unit))
			if not CurePoison_id then
				CurePoison_id = Spell_ID_By_Level('Cleanse',UnitLevel(unit))
			end
			
			if CurePoison_id then
				
				CastSpell(CurePoison_id, SpellBookFrame.bookType)
				if SpellIsTargeting() then SpellTargetUnit(unit) end
				
				return true
				
			end
			
		end
		if ( debuffs['Disease'] ) then
			Debug('CureUnit: Disease Found! Attempting to cure')
			
			CureDisease_id = Spell_ID_By_Level('Abolish Disease',UnitLevel(unit))
			if not CureDisease_id then
				CureDisease_id = Spell_ID_By_Level('Cleanse',UnitLevel(unit))
			end

			if CureDisease_id then
				
				CastSpell(CureDisease_id, SpellBookFrame.bookType)
				if SpellIsTargeting() then SpellTargetUnit(unit) end
				
				return true
				
			end
			
		end
		
		if ( debuffs['Magic'] ) then
			Debug('CureUnit: Magic Found! Attempting to cure')
			CureMagic_id = Spell_ID_By_Level('Dispel Magic',UnitLevel(unit))
			if not CureMagic_id then
				CureMagic_id = Spell_ID_By_Level('Cleanse',UnitLevel(unit))
			end

			if CureMagic_id then
				
				CastSpell(CureMagic_id, SpellBookFrame.bookType)
				if SpellIsTargeting() then SpellTargetUnit(unit) end
				
				return true
				
			end
			
		end
		
	end
	
end

local function UncurseUnit(unit)
	
	if LowestUnitHealth() < min_health_to_uncurse then return end
	
	if UnitCanBeBuffed(unit) 
	and RangeToUnit(unit) < BUFFBOT_EXTREME_RANGE then
		
		if Find_Debuff('Curse', unit) then
			
			Debug('UncurseUnit: Curse Found! Attempting to cure')
			
			RemoveCurse_id = Spell_ID_By_Level("Remove Curse",UnitLevel(unit))
			if not RemoveCurse_id then
				RemoveCurse_id = Spell_ID_By_Level("Remove Lesser Curse",UnitLevel(unit))
			end
			
			if RemoveCurse_id then
				
				CastSpell(RemoveCurse_id, SpellBookFrame.bookType) 
				if SpellIsTargeting() then SpellTargetUnit(unit) end
				
				return true
				
			end
			
		end
		
	end
	
end

local function BuffUnit(unit)
	
	if UnitCanBeBuffed(unit) 
	and RangeToUnit(unit) < BUFFBOT_EXTREME_RANGE then
			
		local buff_id
		local unit_class = UnitClass(unit)
		
		if LowestUnitHealth() < min_health_to_buff then return end
		if not spell_predata[_class] then return end
		
		if  party_ids[unit]
		and not buffs_cast[unit] then
			
			buffs_cast[unit] = {}
			
		end
		
		---------------------------------------------------
		
		-- handle self only buffs
		if unit == 'player' then
			
			for buff,data in pairs(spell_predata[_class]['Self Buffs']) do
			
				if ( data == 1 ) then
					local Time_Since_Buff
					local Buff_Duration = spell_predata[_class][buff]['Duration']
					local Buff_Refresh  = spell_predata[_class][buff]['Refresh']
					
					if buffs_cast[unit][buff] then
						
						Time_Since_Buff = GetTime() - buffs_cast[unit][buff]
						
					else
						
						Time_Since_Buff = 0
						
					end
					
					if (not Find_Buff(buff, unit))
					or ( Buff_Duration - Time_Since_Buff < Buff_Refresh  ) then
						
						buff_id = Spell_ID_By_Level(buff,UnitLevel(unit))
						
						if buff_id then
							
							CastSpell(buff_id, SpellBookFrame.bookType) 
							if SpellIsTargeting() then SpellTargetUnit(unit) end
							
							buffs_cast[unit][buff]= GetTime()
							
							return true
							
						end
						
					end
				end
			end
			
		end
		---------------------------------------------------
		
		-- handle friendly castable buffs
		for buff,data in pairs(spell_predata[_class]['Buffs']) do
			local ok = true;
			if ( data ~= 1 ) then
				ok = false;
			else
				if ( spell_predata[_class][buff]['Classes to Avoid'] ) then
					if ( spell_predata[_class][buff]['Classes to Avoid'][unit_class] ) then
						ok = false;
					end
				end
			end
			if ( ok ) then 
				
				local Time_Since_Buff
				local Buff_Duration = spell_predata[_class][buff]['Duration']
				local Buff_Refresh  = spell_predata[_class][buff]['Refresh']
				
				if  party_ids[unit]
				and buffs_cast[unit][buff] then
					
					Time_Since_Buff = GetTime() - buffs_cast[unit][buff]
					
				else
					
					Time_Since_Buff = 0
					
				end
				
				if (not Find_Buff(buff, unit))
				or ( Buff_Duration - Time_Since_Buff < Buff_Refresh  ) then
					
					buff_id = Spell_ID_By_Level(buff,UnitLevel(unit))
					
					if buff_id then
						
						CastSpell(buff_id, SpellBookFrame.bookType) 
						if SpellIsTargeting() then SpellTargetUnit(unit) end
						
						if party_ids[unit] then
							
							buffs_cast[unit][buff]= GetTime()
							
						end
						
						return true
						
					end
					
				end
				
			end
			
		end
		
	end
	
end

--[[---------------------------------------------------------------------------
BuffBot PUBLIC FUNCTIONS!  - Ok to use these in your macros!!!!
-----------------------------------------------------------------------------]]

function BuffBot()
	
	if UnitCanBeBuffed('target') then
		
		local unit = 'target'
		
		for party_member in party_ids do
			
			if UnitIsUnit('target',party_member) then
				
				unit = party_member
				
			end
			
		end
		
		if CureUnit(unit) then return true end
		if UncurseUnit(unit) then return true end
		if BuffUnit(unit) then return true end
		
	else
		
		for unit in party_ids do
			
			if CureUnit(unit) then return true end
			
		end
		
		for unit in party_ids do
			
			if UncurseUnit(unit) then return true end
			
		end
		
		for unit in party_ids do
			
			if BuffUnit(unit) then return true end
			
		end
		
	end
	
end

function BuffBot_JustBuff()
	
	if UnitCanBeBuffed('target') then
		
		local unit = 'target'
		
		for party_member in party_ids do
			
			if UnitIsUnit('target',party_member) then
				
				unit = party_member
				
			end
			
		end
		
		if BuffUnit(unit) then return true end
		
	else
		
		for unit in party_ids do
			
			if BuffUnit(unit) then return true end
			
		end
		
	end
	
end

function BuffBot_JustCure()
	
	if UnitCanBeBuffed('target') then
		
		if CureUnit('target') then return true end
		if UncurseUnit('target') then return true end
		
	else
		
		for unit in party_ids do
			
			if CureUnit(unit) then return true end
			
		end
		
		for unit in party_ids do
			
			if UncurseUnit(unit) then return true end
			
		end
		
	end
	
end

function BuffBot_JustUncurse()
	
	BigPrint('Archaic function called: BuffBot_JustUncurse()  Please edit your macros!','red')
	--[[
	if UnitCanBeBuffed('target') then
		
		if UncurseUnit('target') then return true end
		
	else
		
		for unit in party_ids do
			
			if UncurseUnit(unit) then return true end
			
		end
		
	end
	--]]
	
end

function BuffBot_Redo()
	
	local Reset = false
	
	for unit in party_ids do
		
		if UnitCanBeBuffed(unit) then
			
			for buff in spell_predata[_class]['Buffs'] do
				
				if not buffs_cast[unit] then buffs_cast[unit] = {} end
				
				if spell_ids[buff] then
					
					buffs_cast[unit][buff] = GetTime() - spell_predata[_class][buff]['Duration']
					Reset = true
					
				end
				
			end
			
			if unit == 'player' then
				
				for buff in spell_predata[_class]['Self Buffs'] do
					
					if not buffs_cast[unit] then buffs_cast[unit] = {} end
					
					if spell_ids[buff] then
						
						buffs_cast[unit][buff] = GetTime() - spell_predata[_class][buff]['Duration']
						Reset = true
						
					end
					
				end
				
			end
			
		end
		
	end
	
	if Reset then BigPrint('Buff timers cleared','red') end
	
	return Reset
	
end

--[[---------------------------------------------------------------------------
Event handler system
-----------------------------------------------------------------------------]]

function BuffBot_OnLoad()
	
	this:RegisterEvent("PLAYER_ENTERING_WORLD")
	this:RegisterEvent("SPELLS_CHANGED")
	this:RegisterEvent("UNIT_AURA")
	this:RegisterEvent("PARTY_MEMBERS_CHANGED")
	
	BINDING_HEADER_BuffBot               = 'BuffBot'
	BINDING_NAME_BuffBot                 = 'Do everything'
	BINDING_NAME_JustBuffs               = 'Just buff'
	BINDING_NAME_JustCure                = 'Just cure debuffs'
	BINDING_NAME_RedoTemps               = 'Force full group rebuff'
	
	Print('You gain BuffBot AddOn!')
	
end

function _OnEvent.PLAYER_ENTERING_WORLD()
	
	_class = UnitClass('player')
	
	_OnEvent.SPELLS_CHANGED()
	
	Load_Party_Buffs()
	
end

function _OnEvent.PARTY_MEMBERS_CHANGED()
	
	Load_Party_Buffs()
	
end

function _OnEvent.SPELLS_CHANGED()
	
	if not _class then return end
	
	Load_Spell_Ids()
	
end

function _OnEvent.UNIT_AURA()
	
	if UnitInParty(arg1) then
		
		Load_Unit_Buffs(arg1)
		Detect_Debuffs()
		
	end
	
end

function BuffBot_OnEvent()
	
	_OnEvent[event]()
	
end
